# Author: Luis Iracheta
# Artificial intelligence engineering
# Universidad Iberoamericana León
import numpy as np
import pandas as pd


# *****************************Class of algoritm cuantitative for rank***********************************
class cl_alg_quantum():
    def __init__(self, funtion, population, num_qubits, cant_ciclos,
                 mutation_percent, i_min,
                    i_max, optimum):
        self.funtion = funtion
        self.population = population
        self.num_qubits = num_qubits
        self.cant_ciclos = cant_ciclos
        self.mutation_percent = mutation_percent
        self.i_min = i_min
        self.i_max = i_max
        self.optimum = optimum


    def run(self):
        print(f"\n[INFO] Starting Quantum algorithm")
        self.quantum_amplitudes_population = self.quantum_amplitudes(self.population, self.num_qubits)
        self.best_solution = None
        self.best_fitness = float('-inf')

        for i in range(self.cant_ciclos):
            self.solutions = self.generate_solutions(self.quantum_amplitudes_population)
            self.fitness_population = self.fitness_quantum_population(self.solutions, self.i_min, self.i_max)

            max_fitness_index = np.argmax(self.fitness_population)
            if self.fitness_population[max_fitness_index] > self.best_fitness:
                self.best_fitness = self.fitness_population[max_fitness_index]
                self.best_solution = self.solutions[max_fitness_index]

            self.quantum_amplitudes_population = self.rotacionGate(self.quantum_amplitudes_population, self.solutions, self.best_solution)
            self.quantum_amplitudes_population = self.mutations_quantum_population(self.quantum_amplitudes_population, self.mutation_percent)
        
        best_population_decimal = self.decode_quantum_population(self.best_solution.reshape(1, -1), self.i_min, self.i_max)
        best_individual = best_population_decimal[0] 
        print(f"[INFO] Best solution found: {best_individual} with fitness: {self.best_fitness}")
        print(f"[INFO] Ending Quantum algorithm\n")   
        return best_individual


    def quantum_amplitudes(self, population, num_qubits):
        amplitudes = np.full((population, num_qubits), 1 / np.sqrt(2))
        return amplitudes

    # Generate solutions to evaluate the quantum population to collapse the qubits into binary values
    def generate_solutions(self, amplitudes):
            [r, c] = amplitudes.shape
            matrix_binary = np.zeros((r,c), dtype=int)
            matrix_random = np.random.rand(r,c)
            
            # Collapse the qubits
            for i in range(r):
                for j in range(c):
                    if matrix_random[i,j] < amplitudes[i,j]:
                        matrix_binary[i,j] = 1
                    else:
                        matrix_binary[i, j] = 0
            
            return matrix_binary
    
    def decode_quantum_population(self, P, Imin, Imax):
        # P: population binary
        # Imin: min interval
        # Imax: max interval
        [r, c] = P.shape # r: number of individuals, c: number of genes
        decimal = np.zeros(r)
        rescaled_decimal = np.zeros(r)

        for i in range(r):
            for j in range(c):
                # Transform from binary to integer decimal
                decimal[i] = decimal[i] + P[i, j] * 2 ** (c - j - 1)
                # Rescale the decimal value in the search space (0 to 2)
                rescaled_decimal[i] = (Imax - Imin) * decimal[i] / (2 ** c - 1) + Imin
        return rescaled_decimal

    def fitness_quantum_population(self, population, Imin, Imax):
        x = self.decode_quantum_population(population, Imin, Imax)
        fitness = self.funtion(x)
        return fitness
    

    # Evolution operators for quantum algorithm
    def rotacionGate(self, P, soluciones, mejorSolucion, theta = 0.1):
        [r, c] = P.shape
        
        for i in range(r):
            for j in range(c):
                if soluciones[i, j] != mejorSolucion[j]:
                    if mejorSolucion[j] == 1:
                        P[i, j] = P[i, j] + np.sin(theta) * (1 - P[i, j])
                    else:
                        P[i, j] = P[i, j] - np.sin(theta) * P[i, j]
        
        P = np.clip(P, 0.05, 0.95)
        return P
    
    def mutations_quantum_population(self, population, mutation_percent):
        [r, c] = population.shape
        matrixAux = np.random.rand(r,c)
        mascara = np.zeros((r, c), dtype=bool)

        for i in range(r):
            for j in range(c):
                if matrixAux[i,j] < mutation_percent:
                    mascara[i,j] = True
                else:
                    mascara[i,j] = False
        
        population = np.where(mascara, 1 - population, population)
        return population

    






'''
Description:

En este archivo se econtraran diferentes modelos genéticos, de manera que se puedan utilizar en el proyecto
de manera sencilla y rápida.

Se agregarán más modelos conforme se vayan necesitando.
'''

import numpy as np
import pandas as pd

# Import the algorithms to be used
from .algorithms.alg_bin import cl_alg_stn_bin_rank
from .algorithms.alg_quantum import cl_alg_quantum


class GEN():
    def __init__(self, funtion, population, cant_genes = None, num_cycles= None, selection_percent = None, 
                 crossing = None, mutation_percent = None, i_min = None, i_max = None, optimum = None, num_qubits = None):
        self.funtion = funtion
        self.population = population
        self.cant_genes = cant_genes
        self.num_qubits = num_qubits
        self.cant_ciclos = num_cycles
        self.selection_percent = selection_percent
        self.crossing = crossing
        self.mutation_percent = mutation_percent
        self.i_min = i_min
        self.i_max = i_max
        self.optimum = optimum 

    def alg_stn_bin_rank(self):
        algoritmo = cl_alg_stn_bin_rank(
            self.funtion,
            self.population,
            self.cant_genes,
            self.cant_ciclos,
            self.selection_percent,
            self.crossing,
            self.mutation_percent,
            self.i_min,
            self.i_max,
            self.optimum
        )
        return algoritmo.run()

    def alg_quantum(self):
        algoritmo = cl_alg_quantum(
            self.funtion,
            self.population,
            self.num_qubits,
            self.cant_ciclos,
            self.mutation_percent,
            self.i_min,
            self.i_max,
            self.optimum
        )
        return algoritmo.run()





# Author: Luis Iracheta
# Artificial intelligence engineering
# Universidad Iberoamericana León
import numpy as np
import pandas as pd


#*****************************Class of algoritm stand binary for rank***********************************
class cl_alg_stn_bin_rank():
    def __init__(self, funtion, population, cant_genes, cant_ciclos, 
                 selection_percent, crossing, mutation_percent, i_min,
                    i_max, optimum):
        self.funtion = funtion
        self.population = population
        self.cant_genes = cant_genes
        self.cant_ciclos = cant_ciclos
        self.selection_percent = selection_percent
        self.crossing = crossing
        self.mutation_percent = mutation_percent
        self.i_min = i_min
        self.i_max = i_max
        self.optimum = optimum

    def run(self):
        print(f"\n[INFO] Starting algorithm: Standard Binary for Rank")
        self.bin_population = self.create_binary_population(self.population, self.cant_genes)
        for i in range(self.cant_ciclos):
            self.Crossing_bin = self.crossing_binary_population(self.bin_population, self.crossing)
            self.Mutations_bin = self.mutations_binary_population(self.Crossing_bin, self.mutation_percent)

            self.select_bin = self.seleccionRanking(self.Mutations_bin, self.selection_percent, self.i_min, self.i_max, self.optimum)

            self.bin_population = self.select_bin.copy()

        best_population_decimal = self.decode_binary_population(self.bin_population, self.i_min, self.i_max)
        best_individual = best_population_decimal[0]
        print(f"[INFO] Best solution found: {best_individual} with fitness: {self.fitness_binary_population(self.bin_population, self.i_min, self.i_max).max()}")
        print(f"[INFO] Ending algoritm stand binary for rank\n")
        return best_individual


    def create_binary_population(self, n, l):
        # n: número de individuos
        # l: longitud de la representación binaria
        self.bin_population = np.random.randint(0, 2, size=(n, l))
        return self.bin_population

    def decode_binary_population(self, P, Imin, Imax):
        # P: population binary
        # Imin: min interval
        # Imax: max interval
        [r, c] = P.shape # r: number of individuals, c: number of genes
        decimal = np.zeros(r)
        rescaled_decimal = np.zeros(r)

        for i in range(r):
            for j in range(c):
                # Transform from binary to integer decimal
                decimal[i] = decimal[i] + P[i, j] * 2 ** (c - j - 1)
                # Rescale the decimal value in the search space (0 to 2)
                rescaled_decimal[i] = (Imax - Imin) * decimal[i] / (2 ** c - 1) + Imin
        return rescaled_decimal

    def fitness_binary_population(self, population, Imin, Imax):
        # [r, c] = population.shape
        x = self.decode_binary_population(population, Imin, Imax)
        fitness = self.funtion(x)
        return fitness
    

    def crossing_binary_population(self, population, cross_percent):
        [r, c] = population.shape
        # cross_percent: percent of population to cross, between 0 and 0.5
        cross_percent = float(cross_percent)
        num_cross = int(r * cross_percent)
        AuxMatrix = np.zeros((2 * num_cross, c))

        for i in range(num_cross):
            r1 = np.random.randint(0, r, size=(1, 2)) # Select two random parents

            # Select the parents to cross
            father1 = population[r1[0, 0], :]
            father2 = population[r1[0, 1], :]

            # Select the crossing point
            r2 = np.random.randint(0, c)

            # Create the children
            children1 = np.concatenate((father1[0:r2], father2[r2:]))
            children2 = np.concatenate((father2[0:r2], father1[r2:]))

            # Save the children in the auxiliary matrix
            AuxMatrix[2 * i, :] = children1
            AuxMatrix[2 * i + 1, :] = children2
        return AuxMatrix


    def mutations_binary_population(self, population, mutation_percent):
        # population: binary population
        # mutation_percent: percent of genes to mutate, between 0 and 0.5
        [r, c] = population.shape
        n = int(mutation_percent * c * r) # Number of genes to mutate
        for i in range(n):
            r1 = np.random.randint(0, r) # Number aleatory to select the individual to mutate
            r2 = np.random.randint(0, c) # Number aleatory to select the gene to mutate
            
            # Compare the value of the gene and change it
            if (population[r1, r2] == 0):
                population[r1, r2] = 1
            else:
                population[r1, r2] = 0

        return population
    
    def seleccionRanking(self, poblacion, select_percent, Imin, Imax, optimum='max'):
        [r, c] = poblacion.shape
        pnew = np.zeros((r, c))  # Matrix to save the new population
        n = int(select_percent * r)  # Number of individuals to select

        fitness = self.fitness_binary_population(poblacion, Imin, Imax).reshape(r, 1)  # Se agrega una columna para guardar su valor Fitness
        expanded_population = np.concatenate([poblacion, fitness], axis=1)  # axis = 0 -> son los renglones, 1 son las columnas

        if (optimum == 'max'):
            indices = np.argsort(expanded_population[:, -1])[::-1]
        elif (optimum == 'min'):
            indices = np.argsort(expanded_population[:, -1])
        else:
            raise ValueError("The optimum parameter must be 'max' or 'min'")
        # Le indicamos con la función argsort que ordene nuesta poblacion, pero como ordena de menos a mayor
        # agregamos el [::1] para que invierta el orden

        organized_population = expanded_population[indices] 
        # Matriz apartir de los indices que ya tenemos

        select_organized_population = organized_population[0:n, :]
        # Seleccionamos hasta el limite establecido "n"

        cleaned_population = select_organized_population[:,0:c]
        # Quitamos el indice de aptitud para quedarnos solo con la población

        for i in range(r):
            for j in range(c):
                if (i < n):
                    pnew[i,j] = cleaned_population[i,j]
                    # Insertamos la población nueva a la matriz de selección
                else:
                    pnew[i, j] = np.random.randint(0,2)
                    # Cuando se acaba, rellenamos con numeros aleatorios
        return pnew



from .gen import GEN
from .algorithms.alg_bin import cl_alg_stn_bin_rank
from .algorithms.alg_quantum import cl_alg_quantum

class RelaxGEN(GEN):
    def __init__(self, funtion=None, population=None, **kwargs):
        super().__init__(funtion, population, **kwargs)
        # Almacena los parámetros específicos como atributos directos
        self.cant_genes = kwargs.get("cant_genes")
        self.num_cycles = kwargs.get("num_cycles")
        self.selection_percent = kwargs.get("selection_percent")
        self.crossing = kwargs.get("crossing")
        self.mutation_percent = kwargs.get("mutation_percent")
        self.i_min = kwargs.get("i_min")
        self.i_max = kwargs.get("i_max")
        self.optimum = kwargs.get("optimum")
        self.num_qubits = kwargs.get("num_qubits")
    
    def alg_stn_bin_rank(self):
        algoritmo = cl_alg_stn_bin_rank(
            funtion=self.funtion,
            population=self.population,
            cant_genes=self.cant_genes,
            cant_ciclos=self.num_cycles,
            selection_percent=self.selection_percent,
            crossing=self.crossing,
            mutation_percent=self.mutation_percent,
            i_min=self.i_min,
            i_max=self.i_max,
            optimum=self.optimum
        )
        return algoritmo.run()

    def alg_quantum(self):
        algoritmo = cl_alg_quantum(
            funtion=self.funtion,
            population=self.population,
            num_qubits=self.num_qubits,
            cant_ciclos=self.num_cycles,
            mutation_percent=self.mutation_percent,
            i_min=self.i_min,
            i_max=self.i_max,
            optimum=self.optimum
        )
        return algoritmo.run()
